/*
 * Create a tar archive.
 *
 * Written 25 Aug 1985 by John Gilmore, ihnp4!hoptoad!gnu.
 *
 * @(#)create.c 1.36 11/6/87 Public Domain - gnu
 */

#include <stdio.h>

#include "tar.h"

extern union record *head;		/* Points to current tape header */
extern struct stat hstat;		/* Stat struct corresponding */
extern int head_standard;		/* Tape header is in ANSI format */

extern char	*malloc();
extern char	*strcpy();
extern char	*strncpy();
extern void	bzero();
extern void	bcopy();
#ifndef VMS
extern int	errno;
#endif

extern void print_header();

union record *start_header();
void finish_header();
void finduname();
void findgname();
char *name_next();
static void to_oct();
void dump_file();

static nolinks;			/* Gets set if we run out of RAM */

void
create_archive()
{
	register char	*p;

	open_archive(0);		/* Open for writing */

	while ((p = name_next()) != NULL) {
		dump_file(p, -1);
	}

	write_eot();
	close_archive();
	name_close();
}		

/*
 * Dump a single file.  If it's a directory, recurse.
 * Result is 1 for success, 0 for failure.
 * Sets global "hstat" to stat() output for this file.
 */
void
dump_file(p, curdev)
	char	*p;			/* File name to dump */
	int	curdev;			/* Device our parent dir was on */
{
	union record	*header;
	char type;
	long extno, exttotal, efsize, efleft;
	int len, i;

	/*
	 * Check for -F exclude flag
	 */

	if (f_exclude > 0) {
		if (f_verbose > 2)
			fprintf(stderr, "exclude %s?\n", p);
		len = strlen(p);
		if ((len >= 5 && strcmp(p+len-5, "/SCCS") == 0) ||
		    strcmp(p, "SCCS") == 0)
			return;
		if (f_exclude > 1) {
			if ((len >= 5 && strcmp(p+len-5, "/errs") == 0) ||
			    (len >= 5 && strcmp(p+len-5, "/core") == 0) ||
			    (len >= 6 && strcmp(p+len-6, "/a.out") == 0) ||
			    strcmp(p, "errs") == 0 ||
			    strcmp(p, "core") == 0 ||
			    strcmp(p, "a.out") == 0)
				return;
			if (len > 2 && p[len-2] == '.' && p[len-1] == 'o')
				return;

			if (f_exclude > 2) {

				/* single letter extensions */

				if (len > 2 && p[len-2] == '.') {

				    /* object code archives */

				    if (p[len-1] == 'a')
					return;

				    /* compressed files */

				    if (p[len-1] == 'z' || p[len-1] == 'Z')
					return;
				}

				/* saved files */

				if (len > 0 && p[len-1] == '-')
					return;

				/* test files */

				if (len > 0 && p[len-1] == '+')
					return;

				/* gnu emacs backup files */

				if (len > 0 && p[len-1] == '~')
					return;

				/* c-tree backup files */

				if (len > 2 && p[len-2] == '_' &&
				    p[len-1] == 'o')
					return;

				/* two letter extensions */

				if (len > 3 && p[len-3] == '.') {
				    /* uu encoded files */

				    if (p[len-2] == 'u' && p[len-1] == 'u')
					return;

				    /* postscript files */

				    if (p[len-2] == 'p' && p[len-1] == 's')
					return;
				    if (p[len-2] == 'P' && p[len-1] == 'S')
					return;

				    /* gzip files */

				    if (p[len-2] == 'g' && p[len-1] == 'z')
					return;

				    /* lzip files */

				    if (p[len-2] == 'l' && p[len-1] == 'z')
					return;

				    /* encrypted files */

				    if (p[len-2] == 'c' && p[len-1] == 'r' &&
					((len > 7 &&
					  strcmp(&p[len-7], ".tar.cr") == 0) ||
					 (len > 10 &&
					  strcmp(&p[len-10], ".tar.gz.cr") == 0))) {
					return;
				    }
				}

				/* junk files */

				if (len >= 4) {
				    /* want p[i+1] = first char of base name */
				    i = len - 1;
				    while (i >= 0 && p[i] != '/') i--;
				    if (i + 4 < len) {
					if (p[i+1] == 'j' && p[i+2] == 'u' &&
					    p[i+3] == 'n' && p[i+4] == 'k')
						return;
				    }
				}

				/* three letter extensions */

				if (len > 4 && p[len-4] == '.') {
					/* error output files */

					if (p[len-3] == 'e' && p[len-2] == 'r')
						return;
					if (p[len-3] == 'l' &&
					    p[len-2] == 'o' && p[len-1] == 'g')
						return;
					if (p[len-3] == 't' &&
					    p[len-2] == 'm' && p[len-1] == 'p')
						return;

					/* postscript files */

					if (p[len-3] == 'e' &&
					    p[len-2] == 'p' && p[len-1] == 's')
						return;
					if (p[len-3] == 'E' &&
					    p[len-2] == 'P' && p[len-1] == 'S')
						return;
					if (p[len-3] == 'p' &&
					    p[len-2] == 'd' && p[len-1] == 'f')
						return;
					if (p[len-3] == 'P' &&
					    p[len-2] == 'D' && p[len-1] == 'F')
						return;

					/* MSDOS binaries */

					if (p[len-3] == 'o' &&
					    p[len-2] == 'b' && p[len-1] == 'j')
						return;
					if (p[len-3] == 'e' &&
					    p[len-2] == 'x' && p[len-1] == 'e')
						return;
					if (p[len-3] == 'l' &&
					    p[len-2] == 'i' && p[len-1] == 'b')
						return;
					if (p[len-3] == 'E' &&
					    p[len-2] == 'X' && p[len-1] == 'E')
						return;

					/* SCS binaries */

					if (p[len-3] == 's' &&
					    p[len-2] == 'c' && p[len-1] == 'r')
						return;
					if (p[len-3] == 'x' &&
					    p[len-2] == 'f' && p[len-1] == 'm')
						return;
					if (p[len-3] == 'x' &&
					    p[len-2] == 'c' && p[len-1] == 'l')
						return;
					if (p[len-3] == 'x' &&
					    p[len-2] == 'p' && p[len-1] == 'l')
						return;
					if (p[len-3] == 'x' &&
					    p[len-2] == 'x' && p[len-1] == 'm')
						return;
					if (p[len-3] == 's' &&
					    p[len-2] == 'p' && p[len-1] == 'c')
						return;

					/* database index files */

					if (p[len-3] == 'd' &&
					    p[len-2] == 'a' && p[len-1] == 't')
						return;
					if (p[len-3] == 'i' &&
					    p[len-2] == 'd' && p[len-1] == 'x')
						return;
					if (p[len-3] == 'f' &&
					    p[len-2] == 'i' && p[len-1] == 'l')
						return;

					/* listing files */

					if (p[len-3] == 'l' &&
					    p[len-2] == 'i' && p[len-1] == 's')
						return;
					if (p[len-3] == 'l' &&
					    p[len-2] == 's' && p[len-1] == 't')
						return;
					if (p[len-3] == 'c' &&
					    p[len-2] == 'r' && p[len-1] == 'f')
						return;
					if (p[len-3] == 'm' &&
					    p[len-2] == 'a' && p[len-1] == 'p')
						return;
					if (p[len-3] == 'd' &&
					    p[len-2] == 'm' && p[len-1] == 'p')
						return;

					/* SCS report writer work files */

					if (p[len-3] == 's' &&
					    p[len-2] == 'e' && p[len-1] == 'l')
						return;
					if (p[len-3] == 't' &&
					    p[len-2] == 'a' && p[len-1] == 'g')
						return;

					/* rnf diversions */

					if (p[len-3] == 'd' &&
					    p[len-2] == 'i' && p[len-1] == 'v')
						return;
					if (p[len-3] == 'D' &&
					    p[len-2] == 'I' && p[len-1] == 'V')
						return;

					/* archives and backups */

					if (p[len-3] == 't' &&
					    p[len-2] == 'a' && p[len-1] == 'r')
						return;
					if (p[len-3] == 't' &&
					    p[len-2] == 'a' && p[len-1] == 'z')
						return;
					if (p[len-3] == 't' &&
					    p[len-2] == 'g' && p[len-1] == 'z')
						return;
					if (p[len-3] == 'z' &&
					    p[len-2] == 'i' && p[len-1] == 'p')
						return;
					if (p[len-3] == 'b' &&
					    p[len-2] == 'a' && p[len-1] == 'k')
						return;
					if (p[len-3] == 's' &&
					    p[len-2] == 'a' && p[len-1] == 'v')
						return;
					if (p[len-3] == 'a' &&
					    p[len-2] == 's' && p[len-1] == 'v')
						return;
					if (p[len-3] == 'b' &&
					    p[len-2] == 'z' && p[len-1] == '2')
						return;

					/* PL index files (.fil is above) */

					if (p[len-3] == 'd' &&
					    p[len-2] == 'i' && p[len-1] == 'c')
						return;
					if (p[len-3] == 'd' &&
					    p[len-2] == 'i' && p[len-1] == 's')
						return;
					if (p[len-3] == 'd' &&
					    p[len-2] == 'p' && p[len-1] == 'm')
						return;
					if (p[len-3] == 'f' &&
					    p[len-2] == 'l' && p[len-1] == 'd')
						return;
					if (p[len-3] == 'h' &&
					    p[len-2] == 'o' && p[len-1] == 'l')
						return;
					if (p[len-3] == 'i' &&
					    p[len-2] == 'n' && p[len-1] == 'f')
						return;
					if (p[len-3] == 'i' &&
					    p[len-2] == 'n' && p[len-1] == 'v')
						return;
					if (p[len-3] == 'i' &&
					    p[len-2] == 'n' && p[len-1] == 'x')
						return;
					if (p[len-3] == 'l' &&
					    p[len-2] == 'o' && p[len-1] == 'c')
						return;
					if (p[len-3] == 't' &&
					    p[len-2] == 't' && p[len-1] == 'l')
						return;

					/* image files */

					if (p[len-3] == 'g' &&
					    p[len-2] == 'i' && p[len-1] == 'f')
						return;
					if (p[len-3] == 'G' &&
					    p[len-2] == 'I' && p[len-1] == 'F')
						return;
					if (p[len-3] == 'j' &&
					    p[len-2] == 'p' && p[len-1] == 'g')
						return;
					if (p[len-3] == 'J' &&
					    p[len-2] == 'P' && p[len-1] == 'G')
						return;
					if (p[len-3] == 'p' &&
					    p[len-2] == 'b' && p[len-1] == 'm')
						return;
					if (p[len-3] == 'p' &&
					    p[len-2] == 'g' && p[len-1] == 'm')
						return;
					if (p[len-3] == 'p' &&
					    p[len-2] == 'n' && p[len-1] == 'g')
						return;
					if (p[len-3] == 'p' &&
					    p[len-2] == 'n' && p[len-1] == 'm')
						return;
					if (p[len-3] == 't' &&
					    p[len-2] == 'i' && p[len-1] == 'f')
						return;
					if (p[len-3] == 'T' &&
					    p[len-2] == 'I' && p[len-1] == 'F')
						return;
					if (p[len-3] == 'b' &&
					    p[len-2] == 'm' && p[len-1] == 'p')
						return;

					/* misc large files */

					if (p[len-3] == 'o' &&
					    p[len-2] == 'u' && p[len-1] == 't')
						return;
					if (p[len-3] == 'd' &&
					    p[len-2] == 'b' && p[len-1] == 'g')
						return;
					if (p[len-3] == 'p' &&
					    p[len-2] == 'c' && p[len-1] == 'c')
						return;

					/* kde files */

					if (p[len-3] == 'k' &&
					    p[len-2] == 'd' && p[len-1] == 'e')
						return;

					/* ddd files */

					if (p[len-3] == 'd' &&
					    p[len-2] == 'd' && p[len-1] == 'd')
						return;

					/* document files */

#if 0 /* keep documentation files */
					if (p[len-3] == 'd' &&
					    p[len-2] == 'o' && p[len-1] == 'c')
						return;
#endif
					if (p[len-3] == 'p' &&
					    p[len-2] == 'p' && p[len-1] == 't')
						return;
				}

				/* four letter extensions */

				if (len > 5 && p[len-5] == '.') {

					/* eps files */

					if (p[len-4] == 'e' && p[len-3] == 'p' &&
					    p[len-2] == 's' && p[len-1] == 'f')
						return;
					if (p[len-4] == 'E' && p[len-3] == 'P' &&
					    p[len-2] == 'S' && p[len-1] == 'F')
						return;
					if (p[len-4] == 'e' && p[len-3] == 'p' &&
					    p[len-2] == 's' && p[len-1] == 'p')
						return;
					if (p[len-4] == 'E' && p[len-3] == 'P' &&
					    p[len-2] == 'S' && p[len-1] == 'P')
						return;

					/* image files */

					if (p[len-4] == 'j' && p[len-3] == 'p' &&
					    p[len-2] == 'e' && p[len-1] == 'g')
						return;
					if (p[len-4] == 'J' && p[len-3] == 'P' &&
					    p[len-2] == 'E' && p[len-1] == 'G')
						return;
					if (p[len-4] == 't' && p[len-3] == 'i' &&
					    p[len-2] == 'f' && p[len-1] == 'f')
						return;
					if (p[len-4] == 'T' && p[len-3] == 'I' &&
					    p[len-2] == 'F' && p[len-1] == 'F')
						return;
				}

				/* gimp-X.X files */

				if (len >= 9 && p[len-9] == '.' &&
				    strncmp(&p[len-9], ".gimp-", 6) == 0) {
					return;
				}

				/* netscape files */

				if (len >= 9 && p[len-9] == '.' &&
				    strncmp(&p[len-9], ".netscape", 9) == 0) {
					return;
				}

				/* mozilla files */

				if (len >= 8 && p[len-8] == '.' &&
				    strncmp(&p[len-8], ".mozilla", 8) == 0) {
					return;
				}

				if (len >= 11 && p[len-11] == '.' &&
				    strncmp(&p[len-11], ".fullcircle", 11) == 0) {
					return;
				}

				/* opera files */

				if (len >= 6 && p[len-6] == '.' &&
				    strncmp(&p[len-6], ".opera", 6) == 0) {
					return;
				}

				/* google earth files */

				if (len >= 12 && p[len-12] == '.' &&
				    strncmp(&p[len-12], ".googleearth", 12) == 0) {
					return;
				}

				if (len >= 12 &&
				    strncmp(&p[len-12], "google-earth", 12) == 0) {
					return;
				}

				/* thunderbird files */

				if (len >= 12 && p[len-12] == '.' &&
				    strncmp(&p[len-12], ".thunderbird", 12) == 0) {
					return;
				}

				/* macromedia files */

				if (len >= 11 && p[len-11] == '.' &&
				    strncmp(&p[len-11], ".macromedia", 11) == 0) {
					return;
				}

				/* .enlightenment files */

				if (len >= 14 && p[len-14] == '.' &&
				    strncmp(&p[len-14], ".enlightenment", 14) == 0) {
					return;
				}

				/* .openoffice.org2 files */

				if (len >= 16 && p[len-16] == '.' &&
				    strncmp(&p[len-16], ".openoffice.org2", 16) == 0) {
					return;
				}

				/* .cleanmail files */

				if (len >= 10 && p[len-10] == '.' &&
				    strncmp(&p[len-10], ".cleanmail", 10) == 0) {
					return;
				}

				/* .java files */

				if (len >= 5 && p[len-5] == '.' &&
				    strncmp(&p[len-5], ".java", 5) == 0) {
					return;
				}

				/* gnome files */

				if (len >= 6 && p[len-6] == '.' &&
				    strncmp(&p[len-6], ".gconf", 6) == 0) {
					return;
				}

				if (len >= 7 && p[len-7] == '.' &&
				    strncmp(&p[len-7], ".gconfd", 7) == 0) {
					return;
				}

				if (len >= 6 && p[len-6] == '.' &&
				    strncmp(&p[len-6], ".gnome", 6) == 0) {
					return;
				}

				if (len >= 14 && p[len-14] == '.' &&
				    strncmp(&p[len-14], ".gnome-desktop", 14) == 0) {
					return;
				}

				if (len >= 7 && p[len-7] == '.' &&
				    strncmp(&p[len-7], ".gnome2", 7) == 0) {
					return;
				}

				if (len >= 6 && p[len-6] == '.' &&
				    strncmp(&p[len-6], ".gtkrc", 6) == 0) {
					return;
				}

				if (len >= 13 && p[len-13] == '.' &&
				    strncmp(&p[len-13], ".ICEauthority", 13) == 0) {
					return;
				}

				if (len >= 7 && p[len-7] == '.' &&
				    strncmp(&p[len-7], ".config", 7) == 0) {
					return;
				}

				if (len >= 9 && p[len-9] == '.' &&
				    strncmp(&p[len-9], ".metacity", 9) == 0) {
					return;
				}

				if (len >= 6 && p[len-6] == '.' &&
				    strncmp(&p[len-6], ".pulse", 6) == 0) {
					return;
				}

				if (len >= 7 && p[len-7] == '.' &&
				    strncmp(&p[len-7], ".ccache", 7) == 0) {
					return;
				}

				if (len >= 14 && p[len-14] == '.' &&
				    strncmp(&p[len-14], ".gtk-bookmarks", 14) == 0) {
					return;
				}

				if (len >= 17 && p[len-17] == '.' &&
				    strncmp(&p[len-17], ".gtkrc-1.2-gnome2", 17) == 0) {
					return;
				}

				if (len >= 3 && p[len-3] == '.' &&
				    strncmp(&p[len-3], ".qt", 3) == 0) {
					return;
				}

				if (len >= 6 && p[len-6] == '.' &&
				    strncmp(&p[len-6], ".local", 6) == 0) {
					return;
				}

				if (len >= 9 && p[len-9] == '.' &&
				    strncmp(&p[len-9], ".esd_auth", 9) == 0) {
					return;
				}

				if (len >= 11 && p[len-11] == '.' &&
				    strncmp(&p[len-11], ".fontconfig", 11) == 0) {
					return;
				}

				/* sawfish files */

				if (len >= 8 && p[len-8] == '.' &&
				    strncmp(&p[len-8], ".sawfish", 8) == 0) {
					return;
				}

				/* nautilus files */

				if (len >= 9 && p[len-9] == '.' &&
				    strncmp(&p[len-9], ".nautilus", 9) == 0) {
					return;
				}

				/* RedHat files */

				if (len >= 16 && p[len-16] == '.' &&
				    strncmp(&p[len-16], ".rhn-applet.conf", 16) == 0) {
					return;
				}

				if (len >= 17 && p[len-17] == '.' &&
				    strncmp(&p[len-17], ".rhn-applet.cache", 17) == 0) {
					return;
				}

				/* nsmail files */

				if (len >= 6 && p[len-6] == 'n' &&
				    (len <= 6 || p[len-7] == '/') &&
				    strncmp(&p[len-6], "nsmail", 6) == 0) {
					return;
				}

				/* wine files */

				if (len >= 5 && p[len-5] == '.' &&
				    strncmp(&p[len-5], ".wine", 5) == 0) {
					return;
				}

				/* nomachine files */

				if (len >= 3 && p[len-3] == '.' &&
				    strncmp(&p[len-3], ".nx", 3) == 0) {
					return;
				}

				/* X files */

				if (len >= 11 && p[len-11] == '.' &&
				    strncmp(&p[len-11], ".Xauthority", 11) == 0) {
					return;
				}

				if (len >= 16 && p[len-16] == '.' &&
				    strncmp(&p[len-16], ".xsession-errors", 16) == 0) {
					return;
				}

				if (len >= 19 && p[len-19] == '.' &&
				    strncmp(&p[len-19], ".nvidia-settings-rc", 19) == 0) {
					return;
				}

				/* xine files */

				if (len >= 5 && p[len-5] == '.' &&
				    strncmp(&p[len-5], ".xine", 5) == 0) {
					return;
				}

				/* evolution files */

				if (len >= 10 && p[len-10] == '.' &&
				    strncmp(&p[len-10], ".evolution", 10) == 0) {
					return;
				}

				/* tomboy */

				if (len >= 7 && p[len-7] == '.' &&
				    strncmp(&p[len-7], ".tomboy", 7) == 0) {
					return;
				}

				/* ncftp */

				if (len >= 6 && p[len-6] == '.' &&
				    strncmp(&p[len-6], ".ncftp", 6) == 0) {
					return;
				}

				/* less */

				if (len >= 8 && p[len-8] == '.' &&
				    strncmp(&p[len-8], ".lesshst", 8) == 0) {
					return;
				}

				/* ssh */

				if (len >= 4 && p[len-4] == '.' &&
				    strncmp(&p[len-4], ".ssh", 4) == 0) {
					return;
				}

				/* skype */

				if (len >= 6 && p[len-6] == '.' &&
				    strncmp(&p[len-6], ".Skype", 6) == 0) {
					return;
				}

				/* gkrellm2 */

				if (len >= 9 && p[len-9] == '.' &&
				    strncmp(&p[len-9], ".gkrellm2", 9) == 0) {
					return;
				}

				/* agency9 3d maps */

				if (len >= 8 && p[len-8] == '.' &&
				    strncmp(&p[len-8], ".agency9", 8) == 0) {
					return;
				}

				/* misc */

				if (len >= 15 && p[len-15] == '.' &&
				    strncmp(&p[len-15], ".gstreamer-0.10", 15) == 0) {
					return;
				}

				if (len >= 5 && p[len-5] == '.' &&
				    strncmp(&p[len-5], ".mcop", 5) == 0) {
					return;
				}
			}
		}
	}

	/*
	 * Use stat if following (rather than dumping) 4.2BSD's
	 * symbolic links.  Otherwise, use lstat (which, on non-4.2
	 * systems, is #define'd to stat anyway.
	 */
	if (0 != f_follow_links? stat(p, &hstat): lstat(p, &hstat))
	{
badperror:
		perror(p);
badfile:
		errors++;
		return;
	}

	/*
	 * Check file age against -S since flag
	 *
	 * mtime changed by creat, mknod, pipe, utime, write
	 * ctime also changed by chmod, chown, link, unlink
	 */

	if (f_since > 0L) {
		if (hstat.st_mtime < f_since &&
			(hstat.st_mode & S_IFMT) != S_IFDIR) {
			if (f_verbose > 2)
				fprintf(stderr, "%s old\n", p);
			return;
		}
#if 0
		if (f_verbose > 2) {
			fprintf(stderr, "%s mtime %ld newer than %ld\n",
				p, hstat.st_mtime, f_since);
		}
#endif
	}

	/*
	 * Check to exclude symbolic links and special files
	 */

	if (f_exclude > 2) {
		switch (hstat.st_mode & S_IFMT) {
#ifdef S_IFLNK
		case S_IFLNK: return; /* Symbolic link */
#endif
#ifdef S_IFCHR
		case S_IFCHR: return; /* Character special file */
#endif
#ifdef S_IFBLK
		case S_IFBLK: return; /* Block special file */
#endif
		default: break;
		}
	}

#ifdef MSDOS
	/*
	 * Fixup mode for MSDOS
	 */
	if (hstat.st_mode & 0111) hstat.st_mode |= 0111;
	if (hstat.st_mode & 0222) hstat.st_mode |= 0222;
	if (hstat.st_mode & 0444) hstat.st_mode |= 0444;
#endif

	/*
	 * See if we are crossing from one file system to another,
	 * and avoid doing so if the user only wants to dump one file system.
	 */
	if (f_local_filesys && curdev >= 0 && curdev != hstat.st_dev) {
		annorec(stderr, tar);
		fprintf(stderr,
			"%s: is on a different filesystem; not dumped\n",
			p);
		return;
	}

	/*
	 * Check for multiple links.
	 *
	 * We maintain a list of all such files that we've written so
	 * far.  Any time we see another, we check the list and
	 * avoid dumping the data again if we've done it once already.
	 */
#ifdef HASINOT
	if (hstat.st_nlink > 1) switch (hstat.st_mode & S_IFMT) {
		register struct link	*lp;

	case S_IFREG:			/* Regular file */
#ifdef S_IFCTG
	case S_IFCTG:			/* Contigous file */
#endif
#ifdef S_IFCHR
	case S_IFCHR:			/* Character special file */
#endif

#ifdef S_IFBLK
	case S_IFBLK:			/* Block     special file */
#endif

#ifdef S_IFIFO
	case S_IFIFO:			/* Fifo      special file */
#endif

		/* First quick and dirty.  Hashing, etc later FIXME */
		for (lp = linklist; lp; lp = lp->next) {
			if (lp->ino == hstat.st_ino &&
			    lp->dev == hstat.st_dev) {
				/* We found a link. */
				hstat.st_size = 0;
				header = start_header(p, &hstat);
				if (header == NULL) goto badfile;
				strcpy(header->header.linkname,
					lp->name);
				header->header.linkflag = LF_LINK;
				finish_header(header);
		/* FIXME: Maybe remove from list after all links found? */
				return;		/* We dumped it */
			}
		}

		/* Not found.  Add it to the list of possible links. */
		lp = (struct link *) malloc( (unsigned)
			(strlen(p) + sizeof(struct link) - NAMSIZ));
		if (!lp) {
			if (!nolinks) {
				fprintf(stderr,
	"tar: no memory for links, they will be dumped as separate files\n");
				nolinks++;
			}
		}
		lp->ino = hstat.st_ino;
		lp->dev = hstat.st_dev;
		strcpy(lp->name, p);
		lp->next = linklist;
		linklist = lp;
	}
#endif

	/*
	 * This is not a link to a previously dumped file, so dump it.
	 */
	switch (hstat.st_mode & S_IFMT) {

	case S_IFREG:			/* Regular file */
#ifdef S_IFCTG
	case S_IFCTG:			/* Contigous file */
#endif
	{
		int	f;		/* File descriptor */
		int	bufsize, count;
		register long	sizeleft, volleft;
		register union record 	*start;
#ifdef	X_MAGIC
		struct xexec xhdr;
#endif
#if	defined(U802WRMAGIC) || defined(ISCOFF)
		unsigned short maghdr;
#endif
#ifdef	ELFMAG
		unsigned char e_mag_ident[SELFMAG];
#endif

		sizeleft = hstat.st_size;
		/* Don't bother opening empty, world readable files. */
		if (sizeleft > 0 || 0444 != (0444 & hstat.st_mode)) {
			f = open(p, O_RDONLY|O_BINARY);
			if (f < 0) goto badperror;
#ifdef	ELFMAG
			if (f_exclude > 2 && (hstat.st_mode & 0111) != 0) {
				if (read(f, e_mag_ident, SELFMAG) == SELFMAG) {
					if (e_mag_ident[0] == ELFMAG0 &&
					    e_mag_ident[1] == ELFMAG1 &&
					    e_mag_ident[2] == ELFMAG2 &&
					    e_mag_ident[3] == ELFMAG3) {
						close(f);
						return;
					}
#if !defined(ISCOFF)
					if (e_mag_ident[0] == 0x4c &&
					    e_mag_ident[1] == 0x01) {
						close(f);
						return;
					}
#endif
				}
				lseek(f, 0L, 0);
			}
#endif
#ifdef	X_MAGIC
			if (f_exclude > 2 && (hstat.st_mode & 0111) != 0) {
				if (read(f, &xhdr, sizeof(xhdr)) ==
					sizeof(xhdr) &&
				    xhdr.x_magic == X_MAGIC) {
					close(f);
					return;
				}
				lseek(f, 0L, 0);
			}
#endif
#if	defined(U802WRMAGIC) || defined(ISCOFF)
			if (f_exclude > 2 && (hstat.st_mode & 0111) != 0) {
				if (read(f, &maghdr, sizeof(maghdr)) ==
					sizeof(maghdr)) {
#if	defined(U802WRMAGIC)
					if (maghdr == U802WRMAGIC ||
					    maghdr == U802ROMAGIC ||
					    maghdr == U802TOCMAGIC) {
						close(f);
						return;
					}
#endif
#if	defined(ISCOFF)
					if (ISCOFF(maghdr)) {
						close(f);
						return;
					}
#endif
				}
				lseek(f, 0L, 0);
			}
#endif
		} else {
			f = -1;
		}

		/* check if extents are needed */
		extno = exttotal = 1L;
		efleft = efsize = hstat.st_size;
		volleft = volumeleft - blocksize +
			(endofrecs()->charptr - ar_record->charptr);
		if (f_oldarch == 2 && ar_size > 0 &&
			volleft < 2L * RECORDSIZE + efsize) {
			if (f_extend == 0 || volleft < 2L * RECORDSIZE) {
				write_eot();
				close_archive();
				diskprompt();
				newdisk();
				volleft = volumeleft - blocksize +
				  (endofrecs()->charptr - ar_record->charptr);
			}
			/* total size               = efsize */
			/* factor to round up       = volumesize - 2RS - 1 */
			/* - amount on current disk = - (volleft - 2RS) */
			exttotal = 1L +
				(efsize + volumesize - volleft - 1L) /
				(volumesize - 2L*RECORDSIZE);
			if (exttotal < 1L) exttotal = 1L;
		}

		header = start_header(p, &hstat);
		if (header == NULL) goto badfile;
#ifdef S_IFCTG
		/* Mark contiguous files, if we support them */
		if (f_standard && (hstat.st_mode & S_IFMT) == S_IFCTG) {
			header->header.linkflag = LF_CONTIG;
		}
#endif

	while (extno <= exttotal) {
		if (exttotal > 1L) {
			if (extno > 1L) {
				write_eot();
				close_archive();
				diskprompt();
				newdisk();
				header = start_header(p, &hstat);
				if (header == NULL) goto badfile;
				sizeleft = volumeleft - 2L*RECORDSIZE;
			} else {
				sizeleft = volleft - 2L*RECORDSIZE;
				fprintf(stdout,
				  "tar: large file %s needs %ld extents.\n",
				   header->header.name, exttotal);
			}
			if (sizeleft > efleft) sizeleft = efleft;
			efleft -= sizeleft;
			fprintf(stdout, " +++ %s %ldK [extent #%ld of %ld]\n",
				header->header.name, (sizeleft+1023L) / 1024L,
				extno, exttotal);
			to_oct(extno, 4, &header->header.magic[0]);
			to_oct(exttotal, 4, &header->header.magic[4]);
			to_oct(efsize, 12, &header->header.uname[0]);
			to_oct(sizeleft, 1+12, header->header.size);
		}
		++extno;
		finish_header(header);
		while (sizeleft > 0) {
			start = findrec();
			bufsize = endofrecs()->charptr - start->charptr;
			if (sizeleft < bufsize) {
				/* Last read -- zero out area beyond */
				bufsize = (int)sizeleft;
				count = bufsize % RECORDSIZE;
				if (count) 
					bzero(start->charptr + sizeleft,
						RECORDSIZE - count);
			}
			count = read(f, start->charptr, bufsize);
			if (count < 0) {
				annorec(stderr, tar);
				fprintf(stderr,
				  "read error at byte %ld, reading %d bytes, in file ",
					hstat.st_size - sizeleft,
					bufsize);
				perror(p);	/* FIXME */
				goto padit;
			}
			sizeleft -= count;
			/* This is nonportable (the type of userec's arg). */
			userec(start+(count-1)/RECORDSIZE);
			if (count == bufsize) continue;
			annorec(stderr, tar);
			fprintf(stderr,
			  "%s: file shrunk by %d bytes, padding with zeros.\n",
				p, sizeleft);
			goto padit;		/* Short read */
		}
	} /* end while for extents */
		if (f >= 0)
			(void)close(f);

		break;

		/*
		 * File shrunk or gave error, pad out tape to match
		 * the size we specified in the header.
		 */
	padit:
		abort();
	}

#ifdef S_IFLNK
	case S_IFLNK:			/* Symbolic link */
	{
		int size;

		hstat.st_size = 0;		/* Force 0 size on symlink */
		header = start_header(p, &hstat);
		if (header == NULL) goto badfile;
		size = readlink(p, header->header.linkname, NAMSIZ);
		if (size < 0) goto badperror;
		if (size == NAMSIZ) {
			annorec(stderr, tar);
			fprintf(stderr,
				"%s: symbolic link too long\n", p);
			break;
		}
		header->header.linkname[size] = '\0';
		header->header.linkflag = LF_SYMLINK;
		finish_header(header);		/* Nothing more to do to it */
	}
		break;
#endif

	case S_IFDIR:			/* Directory */
	{
#ifndef VMS
		register DIR *dirp;
#if USEDIRENT || defined(_AIX)
		register struct dirent *d;
#if USEDIRENT
#define	d_namlen	d_reclen
#endif
#else
		register struct direct *d;
#endif
#endif
		char namebuf[NAMSIZ+2];
		register int len;
		int our_device = hstat.st_dev;

		/* Build new prototype name */
		strncpy(namebuf, p, sizeof (namebuf));
		len = strlen(namebuf);
		while (len >= 1 && '/' == namebuf[len-1]) 
			len--;			/* Delete trailing slashes */
#ifdef MSDOS
		while (len >= 1 && '\\' == namebuf[len-1])
			len--;			/* Delete trailing slashes */
#endif
		namebuf[len++] = '/';		/* Now add exactly one back */
		namebuf[len] = '\0';		/* Make sure null-terminated */

		/*
		 * Output directory header record with permissions
		 * FIXME, do this AFTER files, to avoid R/O dir problems?
		 * If old archive format, don't write record at all.
		 */
		if (!f_oldarch) {
			hstat.st_size = 0;	/* Force 0 size on dir */
			/*
			 * If people could really read standard archives,
			 * this should be:		(FIXME)
			header = start_header(f_standard? p: namebuf, &hstat);
			 * but since they'd interpret LF_DIR records as
			 * regular files, we'd better put the / on the name.
			 */
			header = start_header(namebuf, &hstat);
			if (header == NULL)
				goto badfile;	/* eg name too long */
			if (f_standard) {
				header->header.linkflag = LF_DIR;
			}
#ifdef MSDOS
			if (header->header.name[0])
#endif
			finish_header(header);	/* Done with directory header */
		}

		/* Hack to remove "./" from the front of all the file names */
		if (len == 2 && namebuf[0] == '.') {
			len = 0;
		}

		/* Now output all the files in the directory */
		if (f_dironly)
			break;		/* Unless the user says no */
#ifndef VMS
		errno = 0;
		dirp = opendir(p);
		if (!dirp) {
			if (errno) {
				perror (p);
			} else {
				annorec(stderr, tar);
				fprintf(stderr, "%s: error opening directory",
					p);
			}
			break;
		}
		
		/* Should speed this up by cd-ing into the dir, FIXME */
		while (NULL != (d=readdir(dirp))) {
			/* Skip . and .. */
			if (d->d_name[0] == '.') {
				if (d->d_name[1] == '\0') continue;
				if (d->d_name[1] == '.') {
					if (d->d_name[2] == '\0') continue;
				}
			}
			if (d->d_namlen + len >= NAMSIZ) {
				annorec(stderr, tar);
				fprintf(stderr, "%s%s: name too long\n", 
					namebuf, d->d_name);
				continue;
			}
			strcpy(namebuf+len, d->d_name);
			dump_file(namebuf, our_device);
		}

		closedir(dirp);
#endif
	}
		break;

#ifdef S_IFCHR
	case S_IFCHR:			/* Character special file */
		type = LF_CHR;
		goto easy;
#endif

#ifdef S_IFBLK
	case S_IFBLK:			/* Block     special file */
		type = LF_BLK;
		goto easy;
#endif

#ifdef S_IFIFO
	case S_IFIFO:			/* Fifo      special file */
		type = LF_FIFO;
#endif

	easy:
		if (!f_standard) goto unknown;

		hstat.st_size = 0;		/* Force 0 size */
		header = start_header(p, &hstat);
		if (header == NULL) goto badfile;	/* eg name too long */

		header->header.linkflag = type;
		if (type != LF_FIFO) {
			to_oct((long) major(hstat.st_rdev), 8,
				header->header.devmajor);
			to_oct((long) minor(hstat.st_rdev), 8,
				header->header.devminor);
		}

		finish_header(header);
		break;

	default:
	unknown:
		annorec(stderr, tar);
		fprintf(stderr,
			"%s: Unknown file type; file ignored.\n", p);
		break;
	}
}


/*
 * Make a header block for the file  name  whose stat info is  st .
 * Return header pointer for success, NULL if the name is too long.
 */
union record *
start_header(name, st)
	char	*name;
	register struct stat *st;
{
	register union record *header;
#ifdef MSDOS
	char *n;
#endif

	header = (union record *) findrec();
	bzero(header->charptr, sizeof(*header)); /* XXX speed up */

	/*
	 * Check the file name and put it in the record.
	 */
#ifdef MSDOS
	if (name[0] != '\0' && name[1] == ':') name += 2;
#endif

	if (f_absolute)
	    while ('/' == *name
#ifdef MSDOS
		|| '\\' == *name
#endif
		) {
		static int warned_once = 0;

		name++;				/* Force relative path */
		if (!warned_once++) {
			annorec(stderr, tar);
			fprintf(stderr,
	"Removing leading / from absolute path names in the archive.\n");
		}
	}

	strcpy(header->header.name, name);
#ifdef MSDOS
	/* change to UNIX conventions */
	for(n = header->header.name; *n != '\0'; n++)
		if (*n == '\\') *n = '/';
		else if (*n >= 'A' && *n <= 'Z') *n = *n - 'A' + 'a';
#endif
	if (header->header.name[NAMSIZ-1]) {
		annorec(stderr, tar);
		fprintf(stderr, "%s: name too long\n", name);
		return NULL;
	}

	to_oct((long) (st->st_mode & ~S_IFMT),
					8,  header->header.mode);
	to_oct((long) st->st_uid,	8,  header->header.uid);
	to_oct((long) st->st_gid,	8,  header->header.gid);
	to_oct((long) st->st_size,	1+12, header->header.size);
	to_oct((long) st->st_mtime,	1+12, header->header.mtime);
	/* header->header.linkflag is left as null */

#ifndef NONAMES
	/* Fill in new Unix Standard fields if desired. */
	if (f_standard) {
		header->header.linkflag = LF_NORMAL;	/* New default */
		strcpy(header->header.magic, TMAGIC);	/* Mark as Unix Std */
		finduname(header->header.uname, st->st_uid);
		findgname(header->header.gname, st->st_gid);
	}
#endif
	return header;
}

/* 
 * Finish off a filled-in header block and write it out.
 * We also print the file name and/or full info if verbose is on.
 */
void
finish_header(header)
	register union record *header;
{
	register int	i, sum;
	register char	*p;

	bcopy(CHKBLANKS, header->header.chksum, sizeof(header->header.chksum));

	sum = 0;
	p = header->charptr;
	for (i = sizeof(*header); --i >= 0; ) {
		/*
		 * We can't use unsigned char here because of old compilers,
		 * e.g. V7.
		 */
		sum += 0xFF & *p++;
	}

	/*
	 * Fill in the checksum field.  It's formatted differently
	 * from the other fields:  it has [6] digits, a null, then a
	 * space -- rather than digits, a space, then a null.
	 * We use to_oct then write the null in over to_oct's space.
	 * The final space is already there, from checksumming, and
	 * to_oct doesn't modify it.
	 *
	 * This is a fast way to do:
	 * (void) sprintf(header->header.chksum, "%6o", sum);
	 */
	to_oct((long) sum,	8,  header->header.chksum);
	header->header.chksum[6] = '\0';	/* Zap the space */

	userec(header);

	if (f_verbose) {
		/* These globals are parameters to print_header, sigh */
		head = header;
		/* hstat is already set up */
		head_standard = f_standard;
		print_header(stderr);
	}

	return;
}


/*
 * Quick and dirty octal conversion.
 * Converts long "value" into a "digs"-digit field at "where",
 * including a trailing space and room for a null.  "digs"==3 means
 * 1 digit, a space, and room for a null.
 *
 * We assume the trailing null is already there and don't fill it in.
 * This fact is used by start_header and finish_header, so don't change it!
 *
 * This should be equivalent to:
 *	(void) sprintf(where, "%*lo ", digs-2, value);
 * except that sprintf fills in the trailing null and we don't.
 */
static void
to_oct(value, digs, where)
	register long	value;
	register int	digs;
	register char	*where;
{
	
	--digs;				/* Trailing null slot is left alone */
	where[--digs] = ' ';		/* Put in the space, though */

	/* Produce the digits -- at least one */
	do {
		where[--digs] = '0' + (char)(value & 7); /* one octal digit */
		value >>= 3;
	} while (digs > 0 && value != 0);

	/* Leading spaces, if necessary */
	while (digs > 0)
		where[--digs] = ' ';

}


/*
 * Write the EOT record(s).
 * We actually zero at least one record, through the end of the block.
 * Old tar writes garbage after two zeroed records -- and PDtar used to.
 */
write_eot()
{
	union record *p;
	int bufsize;

	p = findrec();
	bufsize = endofrecs()->charptr - p->charptr;
	if (f_oldarch == 2 && ar_size > 0 && volumeleft > 0L &&
	    ((long) bufsize) > volumeleft)
		bufsize = (int) volumeleft;
	bzero(p->charptr, bufsize);
	userec(p);
}
