/*
 * Diff files from a tar archive.
 *
 * Written 30 April 1987 by John Gilmore, ihnp4!hoptoad!gnu.
 *
 * @(#) diffarch.c 1.10 87/11/11 Public Domain - gnu
 */

#include <stdio.h>

#include "tar.h"

extern int errno;			/* From libc.a */
extern char *valloc();			/* From libc.a */

#ifdef	VMS
extern char vmsname[];			/* temp for making VMS filenames */
#endif

extern union record *head;		/* Points to current tape header */
extern struct stat hstat;		/* Stat struct corresponding */
extern int head_standard;		/* Tape header is in ANSI format */

extern void print_header();
extern void skip_file();

char *filedata;				/* Pointer to area for reading
					   file contents into */

/*
 * Initialize for a diff operation
 */
diff_init()
{

	/*NOSTRICT*/
	filedata = (char *) valloc((unsigned)blocksize);
	if (!filedata) {
		fprintf(stderr,
		"tar: could not allocate memory for diff buffer of %d bytes\n",
			blocking);
		exit(EX_ARGSBAD);
	}
}

/*
 * Diff a file against the archive.
 */
void
diff_archive()
{
	register char *data;
	int fd, check, namelen, written, datain, dataout;
	int err, firsttime;
	long size;
	struct stat filestat;
#ifdef HASINOT
	struct stat linkstat;
#endif
	char linkbuf[NAMSIZ+3];
	char *myname, *mylinkname;
	char shortname[NAMSIZ+1], shortlinkname[NAMSIZ+1];

#ifndef EPIPE
#define	EPIPE	ENOENT
#endif

	errno = EPIPE;			/* FIXME, remove perrors */

	saverec(&head);			/* Make sure it sticks around */
	userec(head);			/* And go past it in the archive */
	decode_header(head, &hstat, &head_standard, 1);	/* Snarf fields */

	/* Print the record from 'head' and 'hstat' */
	if (f_verbose)
		print_header(stdout);

	/* truncate filenames to 14 characters */
	myname = head->header.name;
	if (f_shortnames) {
		if (truncname(myname, shortname))
			fprintf(stdout, "[ %s truncated to %s ]\n",
				myname, shortname);
		myname = shortname;
	}

	/* Check for fully specified pathnames and other atrocities. */

	if (f_absolute)
	    while ('/' == *myname) {
		static int warned_once = 0;

		myname++;			/* Force relative path */
		if (!warned_once++) {
			annorec(stderr, tar);
			fprintf(stderr,
	"Removing leading / from absolute path names in the archive.\n");
		}
	}

	switch (head->header.linkflag) {

	default:
		annofile(stderr, tar);
		fprintf(stderr,
		   "Unknown file type '%c' for %s, diffed as normal file\n",
			head->header.linkflag, myname);
		/* FALL THRU */

	case LF_OLDNORMAL:
	case LF_NORMAL:
	case LF_CONTIG:
		/*
		 * Appears to be a file.
		 * See if it's really a directory.
		 */
		namelen = strlen(myname)-1;
		if (namelen > 0 && myname[namelen] == '/')
			goto really_dir;
#ifdef MSDOS
		if (namelen > 0 && myname[namelen] == '\\')
			goto really_dir;
#endif

#ifdef VMS
		strncpy(vmsname, myname, NAMSIZ-1);
		vmsname[NAMSIZ-1] = '\0';
		vmsfixname(vmsname, 0);
		myname = vmsname;
		fd = open(myname, O_RDONLY | O_BINARY);
#else
		fd = open(myname, O_NDELAY | O_RDONLY | O_BINARY);
#endif

		if (fd < 0) {
			if (errno == ENOENT) {
				/* Expected error -- to stdout */
				annofile(stdout, (char *)NULL);
				fprintf(stdout, "%s: does not exist\n",
					myname);
			} else {
				annofile(stderr, (char *)NULL);
				perror(myname);
			}
			skip_file((long)hstat.st_size);
			goto quit;
		}

		err = fstat(fd, &filestat);
		if (err < 0) {
			annofile(stderr, (char *)NULL);
			fprintf(stderr, "Cannot fstat file ");
			perror(myname);
			skip_file((long)hstat.st_size);
			goto qclose;
		}

		if ((filestat.st_mode & S_IFMT) != S_IFREG) {
			annofile(stdout, (char *)NULL);
			fprintf(stdout, "%s: not a regular file\n",
				myname);
			skip_file((long)hstat.st_size);
			goto qclose;
		}

		filestat.st_mode &= ~S_IFMT;
		if (filestat.st_mode != hstat.st_mode)
			sigh("mode");
		if (filestat.st_uid  != hstat.st_uid)
			sigh("uid");
		if (filestat.st_gid  != hstat.st_gid)
			sigh("gid");
		if (filestat.st_size != hstat.st_size) {
			sigh("size");
			skip_file((long)hstat.st_size);
			goto qclose;
		}
		if (filestat.st_mtime != hstat.st_mtime)
			sigh("mod time");

		firsttime = 0;

		for (size = hstat.st_size;
		     size > 0;
		     size -= written) {
			/*
			 * Locate data, determine max length
			 * writeable, write it, record that
			 * we have used the data, then check
			 * if the write worked.
			 */
			data = findrec()->charptr;
			if (data == NULL) {	/* Check it... */
				annorec(stderr, tar);
				fprintf(stderr, "Unexpected EOF on archive file\n");
				break;
			}
			written = endofrecs()->charptr - data;
			if (written > size) written = size;
			errno = 0;

			if ((f_cr & 1) == 1) {
				datain = dataout = 0;
				while (datain < written) {
					if (filedata[datain] == '\015')
						++datain;
					else ar_crblock[dataout++] =
					     filedata[datain++];
				}
			} else if ((f_cr & 2) == 2) {
				datain = dataout = 0;
				while (datain < written) {
					if (filedata[datain] == '\n')
					  ar_crblock[dataout++] = '\015';
					ar_crblock[dataout++] =
					     filedata[datain++];
				}
			} else {
				dataout = written;
			}
			userec((union record *)(data + written - 1));

			check = read (fd, filedata, dataout);
			/*
			 * The following is in violation of strict
			 * typing, since the arg to userec
			 * should be a struct rec *.  FIXME.
			 */
			if (check == dataout) {
				/* The read worked, now compare the data */
				if (bcmp(data, filedata, check) == 0)
					continue;	/* It compares */
				if (firsttime++ == 0) {
					annofile(stdout, (char *)NULL);
					fprintf(stdout, "%s: data differs\n",
						myname);
				}
			}

			/*
			 * Error in reading from file.
			 * Print it, skip to next file in archive.
			 */
			else {
				annofile(stderr, tar);
				fprintf(stderr,
	"Tried to read %d bytes from file, could only read %d:\n",
					dataout, check);
				perror(myname);
			}
			skip_file((long)(size - written));
			break;	/* Still do the close, mod time, chmod, etc */
		}

	qclose:
		check = close(fd);
		if (check < 0) {
			annofile(stderr, tar);
			fprintf(stderr, "Error while closing ");
			perror(myname);
		}
		
	quit:
		break;

	case LF_LINK:
		mylinkname = head->header.linkname;
		if (f_shortnames) {
			if (truncname(mylinkname, shortlinkname))
				fprintf(stdout, "[ link %s truncated to %s ]\n",
					mylinkname, shortlinkname);
			mylinkname = shortlinkname;
		}
#ifdef HASINOT
		check = 1;
		err = stat(myname, &filestat);
		if (err < 0) {
			annofile(stderr, (char *)NULL);
			fprintf(stderr, "Cannot stat file ");
			perror(myname);
		}
		else {
			err = stat(mylinkname, &linkstat);
			if (err < 0) {
				annofile(stderr, (char *)NULL);
				fprintf(stderr, "Cannot stat file ");
				perror(mylinkname);
			}
			else if (filestat.st_ino == linkstat.st_ino) {
				break;		/* everything ok */
			}
		}
		annofile(stderr, tar);
		fprintf(stderr, "%s: not linked to %s\n",
			myname, mylinkname);
#else

		check = 1;	/* FIXME deal with this */
		/* check = link (mylinkname, myname); */
		/* FIXME, don't worry uid, gid, etc... */
		if (check == 0)
			break;
		annofile(stderr, tar);
		fprintf(stderr, "Could not link %s to ",
			myname);
		perror(mylinkname);
#endif
		break;

#ifdef S_IFLNK
	case LF_SYMLINK:
		check = readlink(myname, linkbuf,
				 (sizeof linkbuf)-1);
		
		if (check < 0) {
			if (errno == ENOENT) {
				annofile(stdout, (char *)NULL);
				fprintf(stdout,
					"%s: no such file or directory\n",
					myname);
			} else {
				annofile(stderr, tar);
				fprintf(stderr, "Could not read link");
				perror(myname);
			}
			break;
		}

		mylinkname = head->header.linkname;
		if (f_shortnames) {
			if (truncname(mylinkname, shortlinkname))
				fprintf(stdout, "[ link %s truncated to %s ]\n",
					mylinkname, shortlinkname);
			mylinkname = shortlinkname;
		}

		linkbuf[check] = '\0';	/* Null-terminate it */
		if (strncmp(mylinkname, linkbuf, check) != 0) {
			annofile(stdout, (char *)NULL);
			fprintf(stdout, "%s: symlink differs\n",
				mylinkname);
		}
		break;
#endif

	case LF_CHR:
		hstat.st_mode |= S_IFCHR;
		goto make_node;

#ifdef S_IFBLK
	/* If local system doesn't support block devices, use default case */
	case LF_BLK:
		hstat.st_mode |= S_IFBLK;
		goto make_node;
#endif

#ifdef S_IFIFO
	/* If local system doesn't support FIFOs, use default case */
	case LF_FIFO:
		hstat.st_mode |= S_IFIFO;
		hstat.st_rdev = 0;		/* FIXME, do we need this? */
		goto make_node;
#endif

	make_node:
#if 1
		err = stat(myname, &filestat);
		if (err < 0) {
			annofile(stderr, (char *)NULL);
			fprintf(stderr, "Cannot stat special file ");
			perror(myname);
		}
		else {
			if ((filestat.st_mode & S_IFMT) !=
			    (hstat.st_mode & S_IFMT)) {
				annofile(stderr, (char *)NULL);
				fprintf(stderr, "%s: not a special file\n",
					myname);
			}
			if ((filestat.st_mode&~S_IFMT) !=
			    (hstat.st_mode&~S_IFMT))
				sigh("mode");
			if (filestat.st_uid  != hstat.st_uid)
				sigh("uid");
			if (filestat.st_gid  != hstat.st_gid)
				sigh("gid");
			if (major(filestat.st_rdev) != major(hstat.st_rdev))
				sigh("major device number");
			if (minor(filestat.st_rdev) != minor(hstat.st_rdev))
				sigh("minor device number");
		}
#else
		/* FIXME, deal with umask */
		
		check = 1; /* FIXME, implement this */
		/* check = mknod(myname, (int) hstat.st_mode,
			(int) hstat.st_rdev); */
		if (check != 0) {
			annofile(stderr, tar);
			fprintf(stderr, "Could not make ");
			perror(myname);
			break;
		};
#endif
		break;

	case LF_DIR:
		/* Check for trailing / */
		namelen = strlen(myname)-1;
	really_dir:
		while (namelen > 0 && myname[namelen] == '/')
			myname[namelen--] = '\0';	/* Zap / */
#ifdef MSDOS
		while (namelen > 0 && myname[namelen] == '\\')
			myname[namelen--] = '\0';	/* Zap / */
#endif

#if 1
		err = stat(myname, &filestat);
		if (err < 0) {
			annofile(stderr, (char *)NULL);
			fprintf(stderr, "Cannot stat directory ");
			perror(myname);
		}
		else {
			if ((filestat.st_mode & S_IFMT) != S_IFDIR) {
				annofile(stderr, (char *)NULL);
				fprintf(stderr, "%s: not a directory\n",
					myname);
			}
			if ((filestat.st_mode&~S_IFMT) !=
			    (hstat.st_mode&~S_IFMT))
				sigh("mode");
			if (filestat.st_uid  != hstat.st_uid)
				sigh("uid");
			if (filestat.st_gid  != hstat.st_gid)
				sigh("gid");
		}
#else
		check = 1; /* FIXME, implement this */
		/* check = mkdir(myname, 0300 | (int)hstat.st_mode); */
		if (check != 0) {
			annofile(stderr, tar);
			fprintf(stderr, "Could not make directory ");
			perror(myname);
			break;
		}
#endif
		
		break;

	}

	/* We don't need to save it any longer. */
	saverec((union record **) 0);	/* Unsave it */
}

/*
 * Sigh about something that differs.
 */
sigh(what)
	char *what;
{

	annofile(stdout, (char *)NULL);
	fprintf(stdout, "%s: %s differs\n",
		head->header.name, what);
}
