/*
 * Header file for public domain tar (tape archive) program.
 *
 * @(#)tar.h 1.24 87/11/06	Public Domain.
 *
 * Created 25 August 1985 by John Gilmore, ihnp4!hoptoad!gnu.
 */

/*
 * Kludge for handling systems that can't cope with multiple
 * external definitions of a variable.  In ONE routine (tar.c),
 * we #define TAR_EXTERN to null; here, we set it to "extern" if
 * it is not already set.
 */
#ifndef TAR_EXTERN
#define TAR_EXTERN extern
#endif

/*
 * System dependent configuration options and header files
 */

#include "port.h"

#include <ctype.h>

#include <errno.h>

#ifdef VMS
# ifdef __VMS_VER
#   if __VMS_VER >= 80000000
#     define VMS8 1
#   endif
# endif
# include <types.h>
# include <stat.h>
# include <file.h>
#else
# include <sys/types.h>
# include <sys/stat.h>
# ifndef V7
#  include <fcntl.h>
#  ifndef MSDOS
#   include <sys/file.h>
#  endif
# endif
#endif

#ifndef	MSDOS
#ifndef	VMS
#include <pwd.h>
#include <grp.h>
#endif
#endif

#if (defined(VMS) && !defined(S_IFMT)) || defined(__ZTC__) || defined(__TURBOC__)
	/* stat.h defines these and also S_IFMT */
typedef unsigned dev_t, ino_t;
#endif

#if defined(m88k) || defined(M_UNIX) || defined(SVR4) || defined(__SVR4) || defined(__linux)
#define	USEDIRENT	1
#else
#define	USEDIRENT	0
#endif

#ifdef USEMSDDIR
#include "msddir.h"
#else
#if USEDIRENT
#include <dirent.h>
#else
#if defined(BSD42) || defined(MSDOS) || defined(_AIX)
#include <sys/dir.h>
#else
/*
 * FIXME: On other systems there is no standard place for the header file
 * for the portable directory access routines.  Change the #include line
 * below to bring it in from wherever it is.
 */
#ifndef VMS
#include "sys/ndir.h"
#endif	/* VMS */
#endif	/* BSD */
#endif	/* DIRENT */
#endif	/* MSDOS */

#include <time.h>		/* get declaration for time() */

#ifdef USG
#include <sys/sysmacros.h>	/* major() and minor() defined here */
#endif

#ifdef	M_XENIX
#ifdef _SCO_DS
#include <sys/x.out.h>		/* defines X_MAGIC for executables */
#else
#include <sys/a.out.h>		/* defines X_MAGIC for executables */
#endif
#endif

#if defined(_AIX) || defined(M_UNIX) || defined(m88k) || defined(SVR4)
#include <filehdr.h>
#endif

#if defined(SVR4)
#define nlist xnlist
#include <sys/x.out.h>
#undef nlist
#include <a.out.h>
#include <elf.h>
#endif

#if defined(__linux)
#include <a.out.h>
#include <elf.h>
#endif

#if defined(__sun__) && (defined(SVR4) || defined(__SVR4))
#include <elf.h>
#endif

#include <signal.h>

/*
 * If there are no symbolic links, there is no lstat().  Use stat().
 */
#ifndef S_IFLNK
#define lstat stat
#endif

/*
 * Some people don't have a #define for these.
 */
#ifndef	O_BINARY
#define	O_BINARY	0
#endif
#ifndef O_NDELAY
#define	O_NDELAY	0
#endif
#ifndef O_RDONLY
#define	O_RDONLY	0
#endif

#ifdef NO_OPEN3
/* We need the #define's even though we don't use them. */
#include "open3.h"
#endif

#ifdef EMUL_OPEN3
/* Simulated 3-argument open for systems that don't have it */
#include "open3.h"
#endif

/*
 * Header block on tape.
 *
 * I'm going to use traditional DP naming conventions here.
 * A "block" is a big chunk of stuff that we do I/O on.
 * A "record" is a piece of info that we care about.
 * Typically many "record"s fit into a "block".
 */
#define	RECORDSIZE	512
#define	NAMSIZ	100
#define	TUNMLEN	32
#define	TGNMLEN	32

union record {
	char		charptr[RECORDSIZE];
	struct header {
		char	name[NAMSIZ];
		char	mode[8];
		char	uid[8];
		char	gid[8];
		char	size[12];
		char	mtime[12];
		char	chksum[8];
		char	linkflag;
		char	linkname[NAMSIZ];
		char	magic[8];
		char	uname[TUNMLEN];
		char	gname[TGNMLEN];
		char	devmajor[8];
		char	devminor[8];
	} header;
};

/* The checksum field is filled with this while the checksum is computed. */
#define	CHKBLANKS	"        "	/* 8 blanks, no null */

/* The magic field is filled with this if uname and gname are valid. */
#define	TMAGIC		"ustar  "	/* 7 chars and a null */

/* The linkflag defines the type of file */
#define	LF_OLDNORMAL	'\0'		/* Normal disk file, Unix compat */
#define	LF_NORMAL	'0'		/* Normal disk file */
#define	LF_LINK		'1'		/* Link to previously dumped file */
#define	LF_SYMLINK	'2'		/* Symbolic link */
#define	LF_CHR		'3'		/* Character special file */
#define	LF_BLK		'4'		/* Block special file */
#define	LF_DIR		'5'		/* Directory */
#define	LF_FIFO		'6'		/* FIFO special file */
#define	LF_CONTIG	'7'		/* Contiguous file */
/* Further link types may be defined later. */

/*
 * Exit codes from the "tar" program
 */
#define	EX_SUCCESS	0		/* success! */
#define	EX_ARGSBAD	1		/* invalid args */
#define	EX_BADFILE	2		/* invalid filename */
#define	EX_BADARCH	3		/* bad archive */
#define	EX_SYSTEM	4		/* system gave unexpected error */


/*
 * Global variables
 */
TAR_EXTERN union record	*ar_block;	/* Start of block of archive */
TAR_EXTERN char		*ar_crblock;	/* Start of cr removal block */
TAR_EXTERN union record	*ar_record;	/* Current record of archive */
TAR_EXTERN union record	*ar_last;	/* Last+1 record of archive block */
TAR_EXTERN char		ar_reading;	/* 0 writing, !0 reading archive */
TAR_EXTERN int		blocking;	/* Size of each block, in records */
TAR_EXTERN int		blocksize;	/* Size of each block, in bytes */
TAR_EXTERN char		*ar_file;	/* File containing archive */
TAR_EXTERN char		*name_file;	/* File containing names to work on */
TAR_EXTERN char		*tar;		/* Name of this program */
TAR_EXTERN int		ar_size;	/* Size of archive media */
TAR_EXTERN long		volumeleft;	/* Bytes remaining on volume */
TAR_EXTERN long		volumesize;	/* Bytes total on volume */

/*
 * Flags from the command line
 */
TAR_EXTERN char	f_absolute;		/* -A */
TAR_EXTERN char	f_reblock;		/* -B */
TAR_EXTERN char	f_create;		/* -c */
TAR_EXTERN char f_cr;			/* -C */
TAR_EXTERN char	f_diff;			/* -d */
TAR_EXTERN char	f_dironly;		/* -D */
TAR_EXTERN char f_extend;		/* -e */
TAR_EXTERN char f_exclude;		/* -F */
TAR_EXTERN char	f_follow_links;		/* -h */
TAR_EXTERN char	f_ignorez;		/* -i */
TAR_EXTERN char	f_keep;			/* -k */
TAR_EXTERN char f_local_filesys;	/* -l */
TAR_EXTERN char	f_modified;		/* -m */
TAR_EXTERN char f_shortnames;		/* -N */
TAR_EXTERN char	f_oldarch;		/* -o */
TAR_EXTERN char	f_use_protection;	/* -p */
TAR_EXTERN char	f_sayblock;		/* -R */
TAR_EXTERN char	f_sorted_names;		/* -s */
TAR_EXTERN long f_since;		/* -S */
TAR_EXTERN char	f_list;			/* -t */
TAR_EXTERN char	f_namefile;		/* -T */
TAR_EXTERN char	f_verbose;		/* -v */
TAR_EXTERN char	f_extract;		/* -x */
TAR_EXTERN char	f_compress;		/* -z, 1=compress, 2=gzip, 3=bzip2 */

/*
 * We now default to Unix Standard format rather than 4.2BSD tar format.
 * The code can actually produce all three:
 *	f_standard	ANSI standard
 *	f_oldarch	V7
 *	neither		4.2BSD
 * but we don't bother, since 4.2BSD can read ANSI standard format anyway.
 * The only advantage to the "neither" option is that we can cmp(1) our
 * output to the output of 4.2BSD tar, for debugging.
 */
#define		f_standard		(!f_oldarch)

/*
 * Structure for keeping track of filenames and lists thereof.
 */
struct name {
	struct name	*next;
	short		length;		/* cached strlen(name) */
	char		found;		/* A matching file has been found */
	char		firstch;	/* First char is literally matched */
	char		regexp;		/* This name is a regexp, not literal */
	char		name[NAMSIZ+1];
};

TAR_EXTERN struct name	*namelist;	/* Points to first name in list */
TAR_EXTERN struct name	*namelast;	/* Points to last name in list */

TAR_EXTERN int		archive;	/* File descriptor for archive file */
TAR_EXTERN int		errors;		/* # of files in error */

/*
 *
 * Due to the next struct declaration, each routine that includes
 * "tar.h" must also include <sys/types.h>.  I tried to make it automatic,
 * but System V has no defines in <sys/types.h>, so there is no way of
 * knowing when it has been included.  In addition, it cannot be included
 * twice, but must be included exactly once.  Argghh!
 *
 * Thanks, typedef.  Thanks, USG.
 */
struct link {
	struct link	*next;
	dev_t		dev;
	ino_t		ino;
	short		linkcount;
	char		name[NAMSIZ+1];
};

TAR_EXTERN struct link	*linklist;	/* Points to first link in list */


/*
 * Error recovery stuff
 */
TAR_EXTERN char		read_error_flag;


/*
 * Declarations of functions available to the world.
 */
union record *findrec();
void userec();
union record *endofrecs();
void anno();
#define	 annorec(stream, msg)	anno(stream, msg, 0)	/* Cur rec */
#define	annofile(stream, msg)	anno(stream, msg, 1)	/* Saved rec */
